

class DatabaseException(Exception):
    def __init__(self, orig):
        self.orig = orig
        super(DatabaseException, self).__init__(orig)

    def __str__(self):
        return 'Orig exception: {}'.format(self.orig)



from sqlalchemy.engine import reflection

from clickhouse_sqlalchemy import Table, engines


class ClickHouseInspector(reflection.Inspector):
    def reflect_table(self, table, *args, **kwargs):
        # This check is necessary to support direct instantiation of
        # `clickhouse_sqlalchemy.Table` and then reflection of it.
        if not isinstance(table, Table):
            table.metadata.remove(table)
            ch_table = Table._make_from_standard(
                table, _extend_on=kwargs.get('_extend_on')
            )
        else:
            ch_table = table

        super(ClickHouseInspector, self).reflect_table(
            ch_table, *args, **kwargs
        )

        with self._operation_context() as conn:
            schema = conn.schema_for_object(ch_table)

            self._reflect_engine(ch_table.name, schema, ch_table)

    def _reflect_engine(self, table_name, schema, table):
        should_reflect = (
            self.dialect.supports_engine_reflection and
            self.dialect.engine_reflection
        )
        if not should_reflect:
            return

        engine_cls_by_name = {e.__name__: e for e in engines.__all__}

        e = self.get_engine(table_name, schema=table.schema)
        if not e:
            raise ValueError("Cannot find engine for table '%s'" % table_name)

        engine_cls = engine_cls_by_name.get(e['engine'])
        if engine_cls is not None:
            engine = engine_cls.reflect(table, **e)
            engine._set_parent(table)
        else:
            table.engine = None

    def get_engine(self, table_name, schema=None, **kw):
        with self._operation_context() as conn:
            return self.dialect.get_engine(
                conn, table_name, schema=schema, info_cache=self.info_cache,
                **kw
            )



from sqlalchemy.sql.base import _generative
from sqlalchemy.sql.selectable import (
    Select as StandardSelect,
    Join
)

from ..ext.clauses import (
    ArrayJoin,
    LeftArrayJoin,
    LimitByClause,
    sample_clause,
)


__all__ = ('Select', 'select')


class Select(StandardSelect):
    _with_cube = False
    _with_rollup = False
    _with_totals = False
    _final_clause = None
    _sample_clause = None
    _limit_by_clause = None
    _array_join = None

    @_generative
    def with_cube(self):
        self._with_cube = True

    @_generative
    def with_rollup(self):
        self._with_rollup = True

    @_generative
    def with_totals(self):
        self._with_totals = True

    @_generative
    def final(self):
        self._final_clause = True

    @_generative
    def sample(self, sample):
        self._sample_clause = sample_clause(sample)

    @_generative
    def limit_by(self, by_clauses, limit, offset=None):
        self._limit_by_clause = LimitByClause(by_clauses, limit, offset)

    def _add_array_join(self, columns, left):
        join_type = ArrayJoin if not left else LeftArrayJoin
        self._array_join = join_type(*columns)

    @_generative
    def array_join(self, *columns, **kwargs):
        left = kwargs.get("left", False)
        self._add_array_join(columns, left=left)

    @_generative
    def left_array_join(self, *columns):
        self._add_array_join(columns, left=True)

    def join(self, right, onclause=None, isouter=False, full=False, type=None,
             strictness=None, distribution=None):
        flags = {
            'full': full,
            'type': type,
            'strictness': strictness,
            'distribution': distribution
        }
        return Join(self, right, onclause=onclause, isouter=isouter,
                    full=flags)


select = Select._create



from sqlalchemy.sql import ClauseElement
from sqlalchemy.sql.schema import ColumnCollectionMixin, SchemaItem, Constraint


class Engine(Constraint):
    __visit_name__ = 'engine'

    def __init__(self, *args, **kwargs):
        pass

    def get_parameters(self):
        return []

    def extend_parameters(self, *params):
        rv = []
        for param in params:
            if isinstance(param, (tuple, list)):
                rv.extend(param)
            elif param is not None:
                rv.append(param)
        return rv

    @property
    def name(self):
        return self.__class__.__name__

    def _set_parent(self, parent, **kwargs):
        self.parent = parent
        parent.engine = self

    @classmethod
    def reflect(cls, table, engine_full, **kwargs):
        raise NotImplementedError


class TableCol(ColumnCollectionMixin, SchemaItem):
    def __init__(self, column, **kwargs):
        super(TableCol, self).__init__(*[column], **kwargs)

    def get_column(self):
        return list(self.columns)[0]


class KeysExpressionOrColumn(ColumnCollectionMixin, SchemaItem):
    def __init__(self, *expressions, **kwargs):
        self.expressions = []

        super(KeysExpressionOrColumn, self).__init__(
            *expressions, _gather_expressions=self.expressions, **kwargs
        )

    def _set_parent(self, table, **kw):
        ColumnCollectionMixin._set_parent(self, table)

        self.table = table

        expressions = self.expressions
        col_expressions = self._col_expressions(table)
        assert len(expressions) == len(col_expressions)
        self.expressions = [
            expr if isinstance(expr, ClauseElement) else colexpr
            for expr, colexpr in zip(expressions, col_expressions)
        ]

    def get_expressions_or_columns(self):
        return self.expressions
